/* GNOME DB libary
 * Copyright (C) 1998,1999 Michael Lausch
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gnome.h>
#include <gda.h>


#include "gda-recordset.h"
#include "gda-command.h"
#include "gda-error.h"

/* per default there's no upper limit on the number of rows which can
 * be fetched from the data source
 */
#define GDA_DEFAULT_MAXROWS 0

/* per default fetch 64 rows at a time */
#define GDA_DEFAULT_CACHESIZE 64



/*
 * this function will change as soon as the client side
 * error objects are here
 */
static gint
Exception(CORBA_Environment *ev)
{
  switch (ev->_major)
    {
    case CORBA_SYSTEM_EXCEPTION:
      fprintf(stderr,"CORBA system exception %s.\n", CORBA_exception_id(ev));
      return -1;
    case CORBA_USER_EXCEPTION:
      fprintf(stderr,"CORBA user exception %s.\n", CORBA_exception_id(ev));
      return -1;
    default:
    }
  return 0;
}


static void
free_chunks(GList* chunks)
{
  GDA_Row* row;
  GList*   ptr;

  ptr = chunks;
  while(ptr)
    {
      row = ptr->data;
      CORBA_free(row);
      ptr = g_list_next(ptr);
    }
}

static GDA_Row*
row_by_idx(Gda_Recordset* rs, gint idx)
{
  GDA_Recordset_Chunk* chunk;

  if (idx < 0 || idx > g_list_length(rs->chunks))
    return 0;
  chunk = g_list_nth(rs->chunks, idx)->data;

  return &chunk->_buffer[0];
}


/**
 * gda_recordset_new:
 *
 * Allocates space for a new recordset.
 *
 * Returns: the allocated recordset object
 */

Gda_Recordset*
gda_recordset_new(void)
{
  Gda_Recordset* rs;

  rs = g_new0(Gda_Recordset, 1);
  rs->maxrows = GDA_DEFAULT_MAXROWS;
  rs->cachesize = GDA_DEFAULT_CACHESIZE;
  rs->readonly = 1;
  rs->chunks_length = 0;
  rs->open = 0;
  rs->eof = rs->bof = 1;
  rs->cursor_location = GDA_USE_CLIENT_CURSOR;
  rs->cursor_type = GDA_OPEN_FWDONLY;
  rs->max_index = -1;
  rs->current_index = -1;
  return rs;
}


/**
 * gda_recordset_free:
 * @rs: the recordset which should be destroyed.
 *
 * This function frees all memory allocated by the recordset and
 * destroys all associations with commands and connections.
 *
 */
void
gda_recordset_free(Gda_Recordset* rs)
{
  if (rs->open)
    gda_recordset_close(rs);
  if (rs->internal_cmd)
    {
      gda_command_free(rs->internal_cmd);
    }
  g_free(rs);
}


/**
 * gda_recordset_close:
 * @rs: the recordset to close
 *
 * This function closes the recordset and frees the memory occupied by 
 * the actual data items. The recordset can be opened
 * again, doing the same query as before.
 * It is guaranteeed that the data cached by the recordset is
 * refetched from the server. Use this function is some
 * characteristics of the recordset must be changed
 *
 */
void
gda_recordset_close(Gda_Recordset* rs)
{
  CORBA_Environment ev;

  CORBA_exception_init(&ev);
  
  if (!rs->open)
    return;
  rs->open = 0;
  if (rs->corba_rs)
    {
      GDA_Recordset_close(rs->corba_rs, &ev);
      Exception(&ev);
    }
  rs->corba_rs = CORBA_OBJECT_NIL;
  if (rs->chunks)
    {
      free_chunks(rs->chunks);
    }
  rs->chunks = 0;
}


/**
 * gda_recordset_bof:
 * @rs: Recordset which should be checked
 *
 * This function is used to check if the recordset cursor is beyond
 * the first row.
 * If this function returns %TRUE any of the functions which actually returns
 * the value of a field an error is returned because the cursor
 * doesn't point to a row. 
 * Returns: 1 if the cursor is beyond the first record or the
 * recordset is empty, 0 otherwise.
 */
gint
gda_recordset_bof(Gda_Recordset* rs)
{
  if (!rs->open || !rs->current_row || rs->bof)
    return 1;
  return 0;
}


/**
 * gda_recordset_eof:
 * @rs:  Recordset which should be checked
 *
 * This function is used to check if the recordset cursor is after
 * the last row. 
 * If this function returns %TRUE any of the functions which actually returns
 * the value of a field an error is returned because the cursor
 * doesn't point to a row. 
 * Returns: 1 if the cursor is after the last record or the
 * recordset is empty, 0 otherwise.
 */
gint
gda_recordset_eof(Gda_Recordset* rs)
{
  if (!rs->open || rs->eof)
    return 1;
  return 0;
}


static gulong
fetch_and_store(Gda_Recordset* rs, gint count, gpointer bookmark)
{
  GDA_Recordset_Chunk* chunk;
  gulong               current_idx;
  CORBA_Environment    ev;
  gint                 current_count;
  
  current_idx = rs->current_index + count;
  if (rs->chunks && current_idx <= rs->max_index && current_idx >= 0)
    {
      /* don't care about direction, it's in our cache */
      rs->bof = rs->eof = 0;
      rs->current_index = current_idx;
      rs->current_row = row_by_idx(rs, rs->current_index);
      return rs->current_index;
    }

  if (count < 0)		
    {
      /* Going backward  and underflow condition */      
      rs->eof = 0;
      rs->bof = 1;
      return rs->current_index;
    }

  rs->eof = 0;			/* be optimistic */
  rs->bof = 0;
  /* going forward. We have to fetch it from the provider */
  CORBA_exception_init(&ev);
  for (current_count = 0; current_count < count; current_count++)
    {
      chunk = GDA_Recordset_fetch(rs->corba_rs, 1, &ev);
      if (Exception(&ev))
	{
	  /* An error */
	  return rs->current_index;
	}
      if (chunk->_length == 0)
	{
	  /* EOF. No more records there */
	  CORBA_free(chunk);
	  rs->current_index = rs->current_index + current_count;
	  rs->max_index = rs->current_index;
	  rs->eof = 1;
	  if (!rs->current_row)
	    rs->bof = 1;
	  return rs->current_index;
	}
      rs->chunks = g_list_append(rs->chunks, chunk);
      if (gda_recordset_eof(rs))
	{
	  current_idx--;
	}
    }
  rs->max_index = current_idx;
  rs->current_index = current_idx;
  rs->current_row = row_by_idx(rs, rs->current_index);
  return rs->current_index;
}
  
static gulong
fetch_and_dont_store(Gda_Recordset* rs, gint count, gpointer bookmark)
{
  GDA_Recordset_Chunk* chunk;
  gulong               current_idx;
  CORBA_Environment    ev;

  current_idx = rs->current_index;
  current_idx += count;
  if (current_idx < 0)
    return GDA_RECORDSET_INVALID_POSITION;
  
  CORBA_exception_init(&ev);
  GDA_Recordset_moveFirst(rs->corba_rs, &ev);
  if (Exception(&ev))
    return GDA_RECORDSET_INVALID_POSITION;
  
  GDA_Recordset_move(rs->corba_rs, current_idx-1, 0, &ev);
  if (Exception(&ev))
    return GDA_RECORDSET_INVALID_POSITION;

  chunk = GDA_Recordset_fetch(rs->corba_rs, count, &ev);
  if (Exception(&ev))
    return GDA_RECORDSET_INVALID_POSITION;

  if (chunk->_length == 0)
    {
      return GDA_RECORDSET_INVALID_POSITION;
    }
  rs->current_index = current_idx;
  if (rs->chunks)
    {
      CORBA_free(rs->chunks->data);
      g_list_free(rs->chunks);
      rs->chunks = 0;
    }
  rs->chunks = g_list_append(rs->chunks, chunk);
  rs->current_row = row_by_idx(rs, rs->current_index);
  return rs->current_index;
}


/**
 * gda_recordset_move:
 * @rs: the recordset
 * @count: The number of records to skip
 * @bookmark: if not NULL, the cursor is positioned relative to the
 * record described by this paramter. seee
 * gda_recordset_get_bookmark() how to get bookmark values for
 * records.
 *
 * Moves the cursor of the recordset forward or backward. @count is
 * the number of records to move. If @count is negative the cursor is
 * moved towards the beginning. The function causes the recordset to
 * actually fetch records from the data source. Each fetch
 * from the data source fetches #cachesize rows in one turn. A maximum 
 * of #maxrows rows can be fetched. 
 *
 * If the cursor is on the second row and the @count parameter is -10, 
 * then the cursor is position in front of the first record
 * available. gda_rcordset_bof() will return %TRUE and 
 * the return value of the function is two, because the cursor actually
 * moved two records. 
 *
 * Returns: the number of the record the cursor is addressing after
 * the move or %GDA_RECORDSET_INVALID_POSITION if there was an error
 * fetching the rows.  
 */
gulong
gda_recordset_move(Gda_Recordset* rs, gint count, gpointer bookmark)
{

  gulong               rc;
  

  g_return_val_if_fail(rs->corba_rs != NULL, GDA_RECORDSET_INVALID_POSITION);
  g_return_val_if_fail(rs->open, GDA_RECORDSET_INVALID_POSITION);
  
  if (count == 0)
    {
      return rs->current_index;
    }

  if (rs->cursor_type == GDA_OPEN_FWDONLY && count < 0)
    return GDA_RECORDSET_INVALID_POSITION;
  
  if (rs->cursor_location == GDA_USE_CLIENT_CURSOR)
    rc = fetch_and_store(rs, count, bookmark);
  else
    rc = fetch_and_dont_store(rs, count, bookmark);
  return rs->current_index;
}


/**
 * gda_recordset_move_first:
 * @rs: the recordset
 *
 * Moves the cursor of the recordset to the first record. 
 * 
 * If the cursor is already on the the first record nothing happen.
 *
 * Returns: the position of the cursor, or
 * %GDA_RECORDSET_INVALID_POSITION
 * if there was an error. 
 */

gulong
gda_recordset_move_first(Gda_Recordset* rs)
{

  g_return_val_if_fail(rs != 0, -1);
  if (rs->cursor_location == GDA_USE_CLIENT_CURSOR)
    {
      GDA_Recordset_Chunk* chunk;
      if (!rs->chunks)
	return gda_recordset_move(rs, 1, 0);
      rs->current_index = 0;
      chunk = rs->chunks->data;
      rs->current_row = &chunk->_buffer[0];
      rs->eof = 0;
      return 0;
    }
  g_error("gda_recordset_move_first for server based cursor NIY");
  return GDA_RECORDSET_INVALID_POSITION;
}


/**
 * gda_recordset_move_last:
 * @rs: the recordset
 *
 * Moves the cursor of the recordset to the last record. 
 * 
 * If the cursor is already on the the last record nothing happen.
 *
 * Returns: the position of the cursor, or
 * %GDA_RECORDSET_INVALID_POSITION
 * if there was an error. 
 */

gulong
gda_recordset_move_last(Gda_Recordset* rs)
{
  gulong pos;
  gulong rc;
  
  if (rs->cursor_location == GDA_USE_CLIENT_CURSOR)
    {
      pos = rs->current_index;
      while (1)
	{
	  rc = gda_recordset_move(rs, 1, 0);
	  
	  if (rc == GDA_RECORDSET_INVALID_POSITION)
	    return GDA_RECORDSET_INVALID_POSITION;
	  if (gda_recordset_eof(rs))
	    return pos;
	  pos++;
	}
    }
  g_error("gda_recordset_move_first for server based cursor NIY");
  return GDA_RECORDSET_INVALID_POSITION;
}


/**
 * gda_recordset_move_next:
 * @rs: the recordset
 *
 * Moves the cursor of the recordset to the next record. 
 * 
 * Has the same effect as calling gda_recordset_move() with @count set to 1.
 *
 * Returns: the position of the cursor, or
 * %GDA_RECORDSET_INVALID_POSITION
 * if there was an error. 
 */

gulong
gda_recordset_move_next(Gda_Recordset* rs)
{

  return gda_recordset_move(rs, 1, 0);
}

/**
 * gda_recordset_field_idx:
 * @rs: the recordset
 * @idx: the index of the field in the current row
 *
 * Returns a pointer to the field at position @idx of
 * the current row of the recordset.
 *
 * Returns: a pointer to a field structor or NULL if 
 * @idx is out of range
 */

Gda_Field*
gda_recordset_field_idx(Gda_Recordset* rs, gint idx)
{

  gint        rowsize;
  Gda_Field*  rc;
  
  g_return_val_if_fail(rs != NULL, 0);
  g_return_val_if_fail(idx >= 0, 0);
  g_return_val_if_fail(rs->open, 0);
  
  if (!rs->current_row && !rs->field_attributes)
    {
      g_warning("This shouldn't happen. Inconsistent recordset\n");
      return 0;
    }
  rowsize = rs->field_attributes->_length;
  if (idx >= rowsize)
    return 0;
  rc = gda_field_new();
  rc->attributes = &rs->field_attributes->_buffer[idx];
  if (rs->current_row)
    {
      rc->real_value =  &rs->current_row->_buffer[idx].realValue;
      rc->shadow_value = &rs->current_row->_buffer[idx].shadowValue;
      rc->original_value = &rs->current_row->_buffer[idx].originalValue;
    }
  return rc;
}

/**
 * gda_recordset_field_name:
 * @rs: the recordset
 * @idx: the name of the field in the current row
 *
 * Returns a pointer to the field with the name  @name of
 * the current row of the recordset.
 *
 * Returns: a pointer to a field structor or NULL if 
 * @idx is out of range
 */

Gda_Field*
gda_recordset_field_name(Gda_Recordset* rs, gchar* name)
{

  gint        rowsize;
  gint        idx;
  
  g_return_val_if_fail(rs != NULL, 0);
  g_return_val_if_fail(rs->open, 0);
  
  if (!rs->current_row && !rs->field_attributes)
    {
      g_warning("This shouldn't happen. Inconsistent recordset\n");
      return 0;
    }
  rowsize = rs->field_attributes->_length;
  for (idx = 0; idx < rowsize; idx++)
    {
      if (strcasecmp(rs->field_attributes->_buffer[idx].name, name) == 0)
	return gda_recordset_field_idx(rs, idx);
    }
  return 0;
}

/**
 * gda_recordset_rowsize:
 *
 * @rs: the recordset
 *
 * Returns the number of fields in a row of the current recordset
 *
 * Returns: the number of rows in the recordset or
 * 0 if there was an error.
 */
gint
gda_recordset_rowsize(Gda_Recordset* rs)
{
  g_return_val_if_fail(rs != NULL, -1);
  g_return_val_if_fail(rs->open, -1);

  if (!rs->current_row && !rs->field_attributes)
    {
      g_warning("This shouldn't happen. Inconsistent recordset\n");
      return 0;
    }
  return rs->field_attributes->_length;
}

/**
 * gda_recordset_set_connection:
 * @rs: the rcordset
 * @cnc: the conneciton
 *
 * Associates a recordset with a connection. This function must not be 
 * called on a already opened  recordst.
 * This function must be called before any of the the
 * gda_recordset_open_txt() or gda_recordset_open_cmd() functions is
 * called.
 *
 * Returns: 0 if everyhting is okay, -1 on error
 */

gint
gda_recordset_set_connection(Gda_Recordset* rs, Gda_Connection* cnc)
{
  if (rs->open)
    return -1;
  rs->cnc = cnc;
  return 0;
}

/**
 * gda_recordset_get_connection:
 * @rs: the rcordset
 *
 * Returns: the connection object used with this recordset. NULL if
 * the recordset isn't open or there's no connection assoctiated with
 * this recordset.
 */
Gda_Connection*
gda_recordset_get_connection(Gda_Recordset* rs)
{
  if (!rs->open)
    return 0;
  return rs->cnc;
}


static void
add_errors_to_list(Gda_Recordset* rs, GDA_DriverError* driver_errors)
{
  gint idx;

  for (idx = 0; idx < driver_errors->errors._length; idx++)
    {
      Gda_Error* e;
      e = gda_error_new();
      e->description = g_strdup(driver_errors->errors._buffer[idx].description);
      e->number      = driver_errors->errors._buffer[idx].number;
      e->source      = g_strdup(driver_errors->errors._buffer[idx].source);
      e->helpurl = 0;
      e->sqlstate    = g_strdup(driver_errors->errors._buffer[idx].sqlstate);
      e->native      = g_strdup(driver_errors->errors._buffer[idx].nativeMsg);
      rs->cnc->errors = g_list_append(rs->cnc->errors, e);
    }
  CORBA_free(driver_errors);
}

/**
 * gda_recordset_open_txt:
 * @rs: the recordset
 * @txt: the command text
 * @cursor_type: the type of the cursor used. Currently only a forward 
 * only cursor is implemented
 * @lock_type: the type of locks used for the command. Currently only
 * readonly type of locking is supported.
 * @options: currently not implemented. Used to identify how to parse
 * the @txt.
 *
 * This function opens a recordsrt. The recordset is filled with the
 * output of @txt. Before this function is called various parameters
 * ofthe recordst might be changed. This is the most featurefull
 * function for retrieving results from a database.
 *
 * Returns: 0 if everytrhing is okay, -1 on error
 */
gint
gda_recordset_open_txt(Gda_Recordset* rs, gchar* txt, GDA_CursorType cursor_type,
		       GDA_LockType lock_type, gulong options)
{
  CORBA_Environment  ev;
  
  if (rs->open)
    return -1;
  rs->cursor_type = cursor_type;
  rs->lock_type   = lock_type;
  CORBA_exception_init(&ev);
  rs->corba_rs = GDA_Connection_openRecordset(rs->cnc->connection, txt,
					      lock_type, cursor_type, options, &ev);
  if (Exception(&ev))
    {
      if (strcmp(CORBA_exception_id(&ev), ex_GDA_DriverError) == 0)
	{
	  add_errors_to_list(rs, (GDA_DriverError*)ev._params);
	}
      if (rs->corba_rs)
	{
	  rs->corba_rs = 0;
	}
    }
  if (!rs->corba_rs)
    return -1;

  rs->field_attributes = GDA_Recordset_describe(rs->corba_rs, &ev);

  if (Exception(&ev) || rs->field_attributes == CORBA_OBJECT_NIL)
    {
      if (strcmp(CORBA_exception_id(&ev), ex_GDA_DriverError) == 0)
	{
	  add_errors_to_list(rs, (GDA_DriverError*)ev._params);
	}
      if (rs->field_attributes)
	{
	  CORBA_free(rs->field_attributes);
	  rs->field_attributes = 0;
	}
      return -1;
    }
  rs->open = 1;
  rs->current_index = -1;
  rs->max_index = -1;
  rs->current_row = 0;
  rs->eof = 1;
  rs->bof = 1;
  return 0;
}

gint
gda_recordset_add_field(Gda_Recordset* rs, Gda_Field* field)
{
  g_return_val_if_fail(rs, -1);
  g_return_val_if_fail(field, -1);
  g_return_val_if_fail(field->attributes, -1);
  g_return_val_if_fail(field->attributes->name, -1);

  if (rs->field_attributes)
    {
      GDA_FieldAttributes* old;
      gint                 idx;

      for (idx = 0; idx < rs->field_attributes->_length; idx++)
	{
	  if (strcasecmp(rs->field_attributes->_buffer[idx].name, field->attributes->name) == 0)
	    return -1;
	}
      rs->field_attributes->_length++;
      old = rs->field_attributes->_buffer;
      rs->field_attributes->_buffer = CORBA_sequence_GDA_FieldAttributes_allocbuf(rs->field_attributes->_length);
      memcpy(rs->field_attributes->_buffer, old, rs->field_attributes->_length - 1);
    }
  else
    {
      rs->field_attributes->_length = 1;
      rs->field_attributes->_buffer = CORBA_sequence_GDA_FieldAttributes_allocbuf(rs->field_attributes->_length);
    }
  rs->field_attributes->_buffer[rs->field_attributes->_length-1].definedSize = field->attributes->definedSize;
  rs->field_attributes->_buffer[rs->field_attributes->_length-1].name = CORBA_string_dup(field->attributes->name);
  rs->field_attributes->_buffer[rs->field_attributes->_length-1].scale = field->attributes->scale;
  rs->field_attributes->_buffer[rs->field_attributes->_length-1].gdaType = field->attributes->gdaType;
  rs->field_attributes->_buffer[rs->field_attributes->_length-1].cType = field->attributes->cType;
  return 0;
}



/**
 * gda_recordset_get_cursorloc:
 * @rs: the recordset
 *
 * Returns: the current value of the cursor location attriburte
 */
GDA_CursorLocation
gda_recordset_get_cursorloc(Gda_Recordset* rs)
{
  return rs->cursor_location;
}

/**
 * gda_recordset_set_cursorloc:
 * @rs: the recordset
 * @loc: the cursor location
 *
 * Set the cursor location attribute to @loc.
 */

void
gda_recordset_set_cursorloc(Gda_Recordset* rs, GDA_CursorLocation loc)
{
  rs->cursor_location = loc;
}

/**
 * gda_recordset_get_cursortype:
 * @rs: the recordset
 *
 * Returns: the current value of the cursor type attriburte
 */
GDA_CursorType
gda_recordset_get_cursortype(Gda_Recordset* rs)
{
  return rs->cursor_type;
}


/**
 * gda_recordset_set_cursortype:
 * @rs: the recordset
 * @type: the cursor type
 *
 * Set the cursor type attribute to @type.
 */
void
gda_recordset_set_cursortype(Gda_Recordset* rs, GDA_CursorType type)
{
  rs->cursor_type = type;
}

